# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the gradients.vjp module."""
from functools import partial

import pytest

import pennylane as qml
from pennylane import numpy as np
from pennylane.gradients import param_shift


class TestComputeVJP:
    """Tests for the numeric computation of VJPs"""

    def test_computation(self):
        """Test that the correct VJP is returned"""
        dy = np.array([[1.0, 2.0], [3.0, 4.0]])
        jac = np.array([[[1.0, 0.1, 0.2], [0.2, 0.6, 0.1]], [[0.4, -0.7, 1.2], [-0.5, -0.6, 0.7]]])

        vjp = qml.gradients.compute_vjp(dy, jac)

        assert vjp.shape == (3,)
        assert np.all(vjp == np.tensordot(dy, jac, axes=[[0, 1], [0, 1]]))

    def test_jacobian_is_none(self):
        """A None Jacobian returns a None VJP"""

        dy = np.array([[1.0, 2.0], [3.0, 4.0]])
        jac = None

        vjp = qml.gradients.compute_vjp(dy, jac)
        assert vjp is None

    def test_zero_dy(self):
        """A zero dy vector will return a zero matrix"""
        dy = np.zeros([2, 2])
        jac = np.array([[[1.0, 0.1, 0.2], [0.2, 0.6, 0.1]], [[0.4, -0.7, 1.2], [-0.5, -0.6, 0.7]]])

        vjp = qml.gradients.compute_vjp(dy, jac)
        assert np.all(vjp == np.zeros([3]))


class TestVJP:
    """Tests for the vjp function"""

    def test_no_trainable_parameters(self):
        """A tape with no trainable parameters will simply return None"""

        with qml.tape.QuantumTape() as tape:
            qml.RX(0.4, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        tape.trainable_params = {}
        dy = np.array([1.0])
        tapes, fn = qml.gradients.vjp(tape, dy, param_shift)

        assert not tapes
        assert fn(tapes) is None

    def test_zero_dy(self):
        """A zero dy vector will return no tapes and a zero matrix"""

        with qml.tape.QuantumTape() as tape:
            qml.RX(0.4, wires=0)
            qml.RX(0.6, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        tape.trainable_params = {0, 1}
        dy = np.array([0.0])
        tapes, fn = qml.gradients.vjp(tape, dy, param_shift)

        assert not tapes
        assert np.all(fn(tapes) == np.zeros([len(tape.trainable_params)]))

    def test_single_expectation_value(self, tol):
        """Tests correct output shape and evaluation for a tape
        with a single expval output"""
        dev = qml.device("default.qubit", wires=2)
        x = 0.543
        y = -0.654

        with qml.tape.JacobianTape() as tape:
            qml.RX(x, wires=[0])
            qml.RY(y, wires=[1])
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0) @ qml.PauliX(1))

        tape.trainable_params = {0, 1}
        dy = np.array([1.0])

        tapes, fn = qml.gradients.vjp(tape, dy, param_shift)
        assert len(tapes) == 4

        res = fn(dev.batch_execute(tapes))
        assert res.shape == (2,)

        expected = np.array([-np.sin(y) * np.sin(x), np.cos(y) * np.cos(x)])
        assert np.allclose(res, expected, atol=tol, rtol=0)

    def test_multiple_expectation_values(self, tol):
        """Tests correct output shape and evaluation for a tape
        with multiple expval outputs"""
        dev = qml.device("default.qubit", wires=2)
        x = 0.543
        y = -0.654

        with qml.tape.JacobianTape() as tape:
            qml.RX(x, wires=[0])
            qml.RY(y, wires=[1])
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))
            qml.expval(qml.PauliX(1))

        tape.trainable_params = {0, 1}
        dy = np.array([1.0, 2.0])

        tapes, fn = qml.gradients.vjp(tape, dy, param_shift)
        assert len(tapes) == 4

        res = fn(dev.batch_execute(tapes))
        assert res.shape == (2,)

        expected = np.array([-np.sin(x), 2 * np.cos(y)])
        assert np.allclose(res, expected, atol=tol, rtol=0)

    def test_prob_expectation_values(self, tol):
        """Tests correct output shape and evaluation for a tape
        with prob and expval outputs"""
        dev = qml.device("default.qubit", wires=2)
        x = 0.543
        y = -0.654

        with qml.tape.JacobianTape() as tape:
            qml.RX(x, wires=[0])
            qml.RY(y, wires=[1])
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))
            qml.probs(wires=[0, 1])

        tape.trainable_params = {0, 1}
        dy = np.array([1.0, 2.0, 3.0, 4.0, 5.0])

        tapes, fn = qml.gradients.vjp(tape, dy, param_shift)
        assert len(tapes) == 4

        res = fn(dev.batch_execute(tapes))
        assert res.shape == (2,)

        expected = (
            np.array(
                [
                    [-2 * np.sin(x), 0],
                    [
                        -(np.cos(y / 2) ** 2 * np.sin(x)),
                        -(np.cos(x / 2) ** 2 * np.sin(y)),
                    ],
                    [
                        -(np.sin(x) * np.sin(y / 2) ** 2),
                        (np.cos(x / 2) ** 2 * np.sin(y)),
                    ],
                    [
                        (np.sin(x) * np.sin(y / 2) ** 2),
                        (np.sin(x / 2) ** 2 * np.sin(y)),
                    ],
                    [
                        (np.cos(y / 2) ** 2 * np.sin(x)),
                        -(np.sin(x / 2) ** 2 * np.sin(y)),
                    ],
                ]
            )
            / 2
        )

        assert np.allclose(res, dy @ expected, atol=tol, rtol=0)


def expected(params):
    x, y = 1.0 * params
    return (
        np.array(
            [
                (np.cos(y / 2) ** 2 * np.sin(x)) + (np.cos(y / 2) ** 2 * np.sin(x)),
                (np.cos(x / 2) ** 2 * np.sin(y)) - (np.sin(x / 2) ** 2 * np.sin(y)),
            ]
        )
        / 2
    )


def ansatz(x, y):
    qml.RX(x, wires=[0])
    qml.RY(y, wires=[1])
    qml.CNOT(wires=[0, 1])
    qml.probs(wires=[0, 1])


class TestVJPGradients:
    """Gradient tests for the vjp function"""

    def test_autograd(self, tol):
        """Tests that the output of the VJP transform
        can be differentiated using autograd."""
        dev = qml.device("default.qubit.autograd", wires=2)
        params = np.array([0.543, -0.654], requires_grad=True)

        def cost_fn(x, dy):
            with qml.tape.JacobianTape() as tape:
                ansatz(x[0], x[1])

            tape.trainable_params = {0, 1}
            tapes, fn = qml.gradients.vjp(tape, dy, param_shift)
            vjp = fn(dev.batch_execute(tapes))
            return vjp

        dy = np.array([-1.0, 0.0, 0.0, 1.0], requires_grad=False)
        res = cost_fn(params, dy)
        assert np.allclose(res, expected(params), atol=tol, rtol=0)

        res = qml.jacobian(cost_fn)(params, dy)
        assert np.allclose(res, qml.jacobian(expected)(params), atol=tol, rtol=0)

    def test_torch(self, tol):
        """Tests that the output of the VJP transform
        can be differentiated using Torch."""
        torch = pytest.importorskip("torch")

        dev = qml.device("default.qubit", wires=2)

        params = torch.tensor([0.543, -0.654], requires_grad=True, dtype=torch.float64)
        dy = torch.tensor([-1.0, 0.0, 0.0, 1.0], dtype=torch.float64)

        with qml.tape.JacobianTape() as tape:
            ansatz(params[0], params[1])

        tape.trainable_params = {0, 1}
        tapes, fn = qml.gradients.vjp(tape, dy, param_shift)
        vjp = fn(qml.execute(tapes, dev, qml.gradients.param_shift, interface="torch"))

        assert np.allclose(vjp.detach(), expected(params.detach()), atol=tol, rtol=0)

        cost = vjp[0]
        cost.backward()

        exp = qml.jacobian(lambda x: expected(x)[0])(params.detach().numpy())
        assert np.allclose(params.grad, exp, atol=tol, rtol=0)

    @pytest.mark.slow
    def test_tf(self, tol):
        """Tests that the output of the VJP transform
        can be differentiated using TF."""
        tf = pytest.importorskip("tensorflow")

        dev = qml.device("default.qubit.tf", wires=2)

        params = tf.Variable([0.543, -0.654], dtype=tf.float64)
        dy = tf.constant([-1.0, 0.0, 0.0, 1.0], dtype=tf.float64)

        with tf.GradientTape() as t:
            with qml.tape.JacobianTape() as tape:
                ansatz(params[0], params[1])

            tape.trainable_params = {0, 1}
            tapes, fn = qml.gradients.vjp(tape, dy, param_shift)
            vjp = fn(dev.batch_execute(tapes))

        assert np.allclose(vjp, expected(params), atol=tol, rtol=0)

        res = t.jacobian(vjp, params)
        assert np.allclose(res, qml.jacobian(expected)(params.numpy()), atol=tol, rtol=0)

    @pytest.mark.slow
    def test_jax(self, tol):
        """Tests that the output of the VJP transform
        can be differentiated using JAX."""
        jax = pytest.importorskip("jax")
        from jax import numpy as jnp

        dev = qml.device("default.qubit.jax", wires=2)
        params = jnp.array([0.543, -0.654])

        @partial(jax.jit, static_argnums=1)
        def cost_fn(x, dy):
            with qml.tape.JacobianTape() as tape:
                ansatz(x[0], x[1])

            tape.trainable_params = {0, 1}
            tapes, fn = qml.gradients.vjp(tape, dy, param_shift)
            vjp = fn(dev.batch_execute(tapes))
            return vjp

        dy = (-1.0, 0.0, 0.0, 1.0)
        res = cost_fn(params, dy)
        assert np.allclose(res, expected(params), atol=tol, rtol=0)

        res = jax.jacobian(cost_fn, argnums=0)(params, dy)
        exp = qml.jacobian(expected)(np.array(params))
        assert np.allclose(res, exp, atol=tol, rtol=0)


class TestBatchVJP:
    """Tests for the batch VJP function"""

    def test_one_tape_no_trainable_parameters(self):
        """A tape with no trainable parameters will simply return None"""
        dev = qml.device("default.qubit", wires=2)

        with qml.tape.QuantumTape() as tape1:
            qml.RX(0.4, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        with qml.tape.JacobianTape() as tape2:
            qml.RX(0.4, wires=0)
            qml.RX(0.6, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        tape1.trainable_params = {}
        tape2.trainable_params = {0, 1}

        tapes = [tape1, tape2]
        dys = [np.array([1.0]), np.array([1.0])]

        v_tapes, fn = qml.gradients.batch_vjp(tapes, dys, param_shift)
        assert len(v_tapes) == 4

        # Even though there are 3 parameters, only two contribute
        # to the VJP, so only 2*2=4 quantum evals
        res = fn(dev.batch_execute(v_tapes))
        assert res[0] is None
        assert res[1] is not None

    def test_all_tapes_no_trainable_parameters(self):
        """If all tapes have no trainable parameters all outputs will be None"""
        dev = qml.device("default.qubit", wires=2)

        with qml.tape.QuantumTape() as tape1:
            qml.RX(0.4, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        with qml.tape.QuantumTape() as tape2:
            qml.RX(0.4, wires=0)
            qml.RX(0.6, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        tape1.trainable_params = set()
        tape2.trainable_params = set()

        tapes = [tape1, tape2]
        dys = [np.array([1.0]), np.array([1.0])]

        v_tapes, fn = qml.gradients.batch_vjp(tapes, dys, param_shift)

        assert v_tapes == []
        assert fn([]) == [None, None]

    def test_zero_dy(self):
        """A zero dy vector will return no tapes and a zero matrix"""
        dev = qml.device("default.qubit", wires=2)

        with qml.tape.QuantumTape() as tape1:
            qml.RX(0.4, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        with qml.tape.JacobianTape() as tape2:
            qml.RX(0.4, wires=0)
            qml.RX(0.6, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        tape1.trainable_params = {0}
        tape2.trainable_params = {0, 1}

        tapes = [tape1, tape2]
        dys = [np.array([0.0]), np.array([1.0])]

        v_tapes, fn = qml.gradients.batch_vjp(tapes, dys, param_shift)
        res = fn(dev.batch_execute(v_tapes))

        # Even though there are 3 parameters, only two contribute
        # to the VJP, so only 2*2=4 quantum evals
        assert len(v_tapes) == 4
        assert np.allclose(res[0], 0)

    def test_reduction_append(self):
        """Test the 'append' reduction strategy"""
        dev = qml.device("default.qubit", wires=2)

        with qml.tape.JacobianTape() as tape1:
            qml.RX(0.4, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        with qml.tape.JacobianTape() as tape2:
            qml.RX(0.4, wires=0)
            qml.RX(0.6, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        tape1.trainable_params = {0}
        tape2.trainable_params = {0, 1}

        tapes = [tape1, tape2]
        dys = [np.array([1.0]), np.array([1.0])]

        v_tapes, fn = qml.gradients.batch_vjp(tapes, dys, param_shift, reduction="append")
        res = fn(dev.batch_execute(v_tapes))

        # Returned VJPs will be appended to a list, one vjp per tape
        assert len(res) == 2
        assert all(isinstance(r, np.ndarray) for r in res)
        assert all(len(r) == len(t.trainable_params) for t, r in zip(tapes, res))

    def test_reduction_extend(self):
        """Test the 'extend' reduction strategy"""
        dev = qml.device("default.qubit", wires=2)

        with qml.tape.JacobianTape() as tape1:
            qml.RX(0.4, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        with qml.tape.JacobianTape() as tape2:
            qml.RX(0.4, wires=0)
            qml.RX(0.6, wires=0)
            qml.CNOT(wires=[0, 1])
            qml.expval(qml.PauliZ(0))

        tape1.trainable_params = {0}
        tape2.trainable_params = {0, 1}

        tapes = [tape1, tape2]
        dys = [np.array([1.0]), np.array([1.0])]

        v_tapes, fn = qml.gradients.batch_vjp(tapes, dys, param_shift, reduction="extend")
        res = fn(dev.batch_execute(v_tapes))

        # Returned VJPs will be extended into a list. Each element of the returned
        # list will correspond to a single input parameter of the combined
        # tapes.
        assert len(res) == sum(len(t.trainable_params) for t in tapes)
